#!/usr/bin/env python

"""
Native library generation.

Copyright (C) 2007, 2008, 2009, 2010, 2011 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.code import Assembler
from micropython.rsvp import *

class NativeLibrary(Assembler):

    "A native library generator providing common routines."

    def __init__(self, program):

        """
        Initialise a native library for the given 'program'.
        """

        Assembler.__init__(self, program)

        self.importer = self.program.get_importer()
        self.builtins = None

        self.unknown_target_block = None
        self.unknown_target_testable_self_block = None
        self.known_target_testable_self_block = None

        # Reset the assembler.

        self.reset()

    def get_native_code(self):
        self.builtins = self.importer.modules.get("__builtins__")

        if self.unknown_target_block is not None:
            self._generateCallFuncUnknownTarget()

        if self.unknown_target_testable_self_block is not None:
            self._generateCallFuncUnknownTargetTestableSelf()

        if self.known_target_testable_self_block is not None:
            self._generateCallFuncKnownTargetTestableSelf()

        return self.blocks

    def _commonCallFunc(self, adjust_block, continue_block):
        self.new_op(CheckContext(working="source", target="status"))
        self.new_op(JumpIfFalse(adjust_block, working="status"))

        # Skip adjustment and tests if the context is not a class.
        # Classes themselves employ a placeholder context so that
        # instantiators can be callable with a context which will be
        # overwritten in the frame.

        # Here, the source value should still refer to the context.

        self.new_op(CheckClass(working="source", target="status"))
        self.new_op(JumpIfFalse(continue_block, working="status"))

    def _commonTestContext(self, success_block):
        self.new_op(CheckInstance(source="source", target="status"))
        self.new_op(JumpIfTrue(success_block, working="status"))

        # Where the context is inappropriate, drop the incomplete frame and
        # raise an exception.

        self.new_op(DropFrame())

        self.make_exception("TypeError")
        self.set_target("exception")
        self.new_op(RaiseException())

    def _commonAdjustFrame(self, adjust_block, continue_block):
        self.set_block(adjust_block)
        self.new_op(AdjustFrame(1))
        self.set_block(continue_block)

    def getCallFuncUnknownTarget(self):
        if self.unknown_target_block is None:
            self.unknown_target_block = self.new_block()

        return self.unknown_target_block

    def _generateCallFuncUnknownTarget(self):

        """
        Some preliminary tests where the target of an invocation is not known.
        Requires the context in the working value register.
        """

        adjust_block = self.new_block()
        continue_block = self.new_block()

        self.set_block(self.unknown_target_block)
        self._commonCallFunc(adjust_block, continue_block)
        self._commonAdjustFrame(adjust_block, continue_block)
        self.new_op(Return())

    def getCallFuncUnknownTargetTestableSelf(self):
        if self.unknown_target_testable_self_block is None:
            self.unknown_target_testable_self_block = self.new_block()

        return self.unknown_target_testable_self_block

    def _generateCallFuncUnknownTargetTestableSelf(self):

        """
        Test any explicit first argument against the context.
        """

        adjust_block = self.new_block()
        continue_block = self.new_block()

        self.set_block(self.unknown_target_testable_self_block)
        self._commonCallFunc(adjust_block, continue_block)
        self._commonTestContext(adjust_block)
        self._commonAdjustFrame(adjust_block, continue_block)
        self.new_op(Return())

    def getCallFuncKnownTargetTestableSelf(self):
        if self.known_target_testable_self_block is None:
            self.known_target_testable_self_block = self.new_block()

        return self.known_target_testable_self_block

    def _generateCallFuncKnownTargetTestableSelf(self):

        """
        Test any explicit first argument against the context.
        """

        adjust_block = self.new_block()
        continue_block = self.new_block()

        self.set_block(self.known_target_testable_self_block)
        self._commonTestContext(continue_block)
        self._commonAdjustFrame(adjust_block, continue_block)
        self.new_op(Return())

# vim: tabstop=4 expandtab shiftwidth=4
